import { Injectable } from '@angular/core';
import { HttpClientModule, HttpClient, HttpHeaders, HttpErrorResponse } from '@angular/common/http';
import { Observable } from 'rxjs/Observable';
import { fromEvent, throwError } from 'rxjs';
import { of } from 'rxjs/observable/of';
import { catchError, map, tap } from 'rxjs/operators';
import { Job } from './job';
import {BehaviorSubject} from 'rxjs/BehaviorSubject';
import { Subject } from 'rxjs/Subject';
import { Strudelapp, StrudelappInstance } from './strudelapp';
import { Computesite } from './computesite';
import { APIServer } from './apiserver';
import { Identity, AuthToken, KeyCert, SshAuthzServer } from './identity';
import { BatchInterface } from './batchinterface';
import { ComputesitesService } from './computesites.service';
import { StrudelappsService } from './strudelapps.service';
import { timer } from 'rxjs/observable/timer';
import { repeat } from 'rxjs/operators';
import { LocationStrategy, Location } from '@angular/common';
import { ActivatedRoute, Router } from '@angular/router';
import { ModaldialogComponent } from './modaldialog/modaldialog.component';
import { MatDialog, MatDialogRef, MAT_DIALOG_DATA, MatDialogModule } from '@angular/material';
import { AuthorisationService } from './authorisation.service';
import { environment } from '../environments/environment';

/** The TES service contains ways to start Tunnels, and Execute programs
Its also responsible for querying a compute site for running jobs */



@Injectable({
  providedIn: 'root',
})
export class TesService {
public Base=environment.tesurl;
public twsproxy = environment.twsurl;
// public Base='http://localhost:5000';
public statusMsg: BehaviorSubject<any>;
public jobs: any[];
public busy: BehaviorSubject<boolean> ;
// public joblist: BehaviorSubject<{ [id: string ]: Job[]}>;
public joblist: BehaviorSubject<Job[]>;

private timerSubscription: any;
private appwindow: any;
private appwindowWatcher: any;
public apiserver: BehaviorSubject<APIServer>;
public apiservers: BehaviorSubject<APIServer[]>;
// public batchinterface: {[id: string] : BatchInterface};

  constructor(private http: HttpClient,
              public dialog: MatDialog,
              private computesitesService: ComputesitesService,
              private authorisationService: AuthorisationService,
              private strudelappsService: StrudelappsService,
              private location: Location ) {

    this.busy = new BehaviorSubject<boolean>(false);
    // this.joblist = new BehaviorSubject<{[id: string]: Job[]}>({});
    this.joblist = new BehaviorSubject<Job[]>([]);
      this.apiserver = new BehaviorSubject<APIServer>(undefined);
    this.apiservers = new BehaviorSubject<APIServer[]>([]);
    this.apiserver.subscribe((r)=>this.logApiServerChange(r));

    this.timerSubscription = null;
    this.appwindowWatcher = null;
    this.getAPIServers();
    // this.batchinterface = {};
    // this.computesitesService.identities.subscribe(identities => this.startPolling(identities));
 }

public setStatusMsg(statusMsg: BehaviorSubject<any>) {
  this.statusMsg = statusMsg;
}

private logApiServerChange(r) {
    console.log('api server was updated');
    console.log(this.apiserver.value);
}

 setApiServer(server: APIServer) {
     console.log('calling setAPIServer');
     this.apiserver.next(server);
 }

 getAPIServers() {
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: false};
   this.http.get<SshAuthzServer[]>('./assets/config/apiservers.json',options)
                    .subscribe(resp => this.updateAPIServers(resp));
 }

 private updateAPIServers(resp) {
     this.apiservers.next(<APIServer[]>resp);
     this.apiserver.next(this.apiservers.value[0]);
 }


 private buildParams(app: Strudelapp, identity: Identity, batchinterface: BatchInterface, appinst?: any): string {
   let params = new URLSearchParams();
   let id = identity.copy_skip_catalog();
   id.site.appCatalog = null;

   if (appinst!== null)  {
     params.set('appinstance',JSON.stringify(appinst));
   }
   params.set('app',JSON.stringify(app));
   params.set('interface',JSON.stringify(batchinterface));
   params.set('identity',JSON.stringify(id));
   return params.toString();
 }

 updateJoblist(resp, identity: Identity) {
   let joblist = <Job[]>resp;
   let alljobs = this.joblist.value;
   let i = 0;
   for (let j of joblist) {
     j.app = this.strudelappsService.getApp(j.name,identity.site.appCatalog.value);
     j.identity = identity;
   }
   this.joblist.next(joblist)
   // alljobs[identity.repr()] = joblist;
   // this.joblist.next(alljobs);
 }

 private getBatchInterfaceError(error: any) {
   console.error(error);
 }

 private getJobsError(error: any) {
   this.joblist.next([]);
   if (error.status == 0) {
     this.statusMsg.next("A network error occured. Please try again latter");
     return
   }
   console.error(error);
   if (error.status == 404 || error.status == 400) {
     this.statusMsg.next("Login expired. Please log in again.");
     this.authorisationService.updateAgentContents();
     return
   }

 }

 getJobs(identity: Identity) {
  let headers = new HttpHeaders();
  let options = { headers: headers, withCredentials: true};
  // remove from the job list any jobs for identities that we don't know about
  let oldjobs = this.joblist.value;
  if ( identity === undefined) {
    this.joblist.next([]);
    return
  }
  let bi = new BatchInterface();
  bi.cancelcmd = identity.site.cancelcmd;
  bi.statcmd = identity.site.statcmd;
  let params = new URLSearchParams();
  params.set('statcmd',JSON.stringify(identity.site.statcmd));
  params.set('host',JSON.stringify(identity.site.host));
  params.set('username',JSON.stringify(identity.username));
  console.log('getjob with params',params.toString());
  
  this.http.get<Job[]>(this.Base+'/stat'+'?'+params.toString(),options)
                // .pipe(catchError(this.networkError))
                .subscribe(resp => this.updateJoblist(resp, identity),
                           error => this.getJobsError(error));
}

 public getconfig(app: Strudelapp, identity: Identity): Observable<any> {
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   return this.http.get<any>(identity.site.url+'getconfig/'+app.name,options)
                                        // .pipe(catchError(this.handleError))
}

 submissionError(error: any) {
   if (error.status != 0) {
     this.statusMsg.next('Job submission failed');
     //this.statusMsg.next(error.error.message);
     console.error(error.error.message);
     this.busy.next(false);
   }
 }

 buildBody(app: Strudelapp, appparams?: string) {
   return JSON.stringify({'app': app, 'appparams': appparams});
 }

 submit(app: Strudelapp, identity: Identity, batchinterface: BatchInterface, appparams?: any) {
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   this.statusMsg.next('Submitting job');
   this.busy.next(true);
   let paramstr = this.buildParams(app,identity,batchinterface);
   // let body = this.buildBody(app,appparams)
   let keys = this.authorisationService.getKeys();
   let body = {'app': app, 'appparams': appparams, 'keys': keys}
   console.log('tes is submitting the job with body',body);
   this.http.post<any>(this.Base+'/submit'+'?'+paramstr, body, options)
                                          .pipe(catchError(this.handleError))
                                          .subscribe(resp => {
                                                              this.busy.next(false);
                                                              this.statusMsg.next(null)
                                                            },
                                                    error => this.submissionError(error));
 }

 submitted(resp: any, identity: Identity ) {
   this.busy.next(false);
   // this.statusMsg.next(null);
   this.getJobs(identity);
 }

 cancel(job: Job) {
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   // this.statusMsg.next(null);
   let data = {};
   let bi = new BatchInterface();
   bi.statcmd = job.identity.site.statcmd;
   bi.cancelcmd = job.identity.site.cancelcmd;
   let paramstr = this.buildParams(job.app,job.identity,bi);
   this.http.delete<any>(this.Base+'/cancel/'+job.jobid+'?'+paramstr, options)
                      .pipe(catchError(this.handleError))
                      .subscribe(resp => this.submitted(resp,job.identity));
 }

 public watchAppwindow(appwindow, dialogRef) {
   if (appwindow.closed) {
     dialogRef.close();
   }
 }

 public getAppInstance(job: Job) {
   let username = job.identity.username;
   let loginhost = job.identity.site.host;
   let batchhost = job.batch_host;
   let params = new URLSearchParams;
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   params.set('cmd',JSON.stringify(job.app.paramscmd));
   let paramstr = params.toString();
   job.connectionState = 1;
   this.http.get<string>(this.Base+'/appinstance/'+username+'/'+loginhost+'/'+batchhost+'?'+paramstr, options)
                // .pipe(catchError(this.handleError))
                .subscribe(resp =>  { job.appinst = resp; this.createTunnel(job) } )
   // let paramstr = this.buildParams(job.app,job.identity,this.batchinterface[job.identity.repr()]);
   // let headers = new HttpHeaders();
   // let options = { headers: headers, withCredentials: true};
   // this.http.get<any>(this.Base+'/getappinst/'+paramstr,options)
   //                                      .pipe(catchError(this.handleError))
   //                                      .subscribe(resp =>  { job.appinst = <StrudelappInstance>resp; this.openAppWindow(job);});
 }

 public createTunnel(job: Job) {
   let username = job.identity.username;
   let loginhost = job.identity.site.host;
   let batchhost = job.batch_host;
   let params = new URLSearchParams;
   let headers = new HttpHeaders();
   job.connectionState = 2;
   let options = { headers: headers, withCredentials: true};
   params.set('cmd',JSON.stringify(job.app.paramscmd));
   let paramstr = params.toString();
   this.http.post<string>(this.Base+'/createtunnel/'+username+'/'+loginhost+'/'+batchhost+'?'+paramstr, job.appinst, options)
                // .pipe(catchError(this.handleError))
                .subscribe(() =>  { this.getAppUrl(job) } )
 }

 public getAppUrl(job: Job) {
   let params = new URLSearchParams;
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   params.set('app',JSON.stringify(job.app));
   params.set('appinst',JSON.stringify(job.appinst));
   let paramstr = params.toString();
   job.connectionState = 3;
   this.http.get<string>(this.Base+'/appurl?'+paramstr,options)
                .pipe(catchError(this.handleError))
                .subscribe(resp => {  job.connectionState = 0; this.openAppWindow(resp)});


 }

 public getInterface(job: Job) {
   this.getAppInstance(job);
 }

 public openAppWindow(url: any) {
     console.log('in open window url is',url,this.twsproxy);
     let windowloc = url.replace(/\{twsproxy\}/g,this.twsproxy);
     console.log('window loc is ',windowloc);
   // let windowloc = this.router.config
   this.appwindow = window.open(windowloc);
   if (!(this.appwindowWatcher === null)) {
     this.appwindowWatcher.unsubscribe()
   }

   let dialogRef = this.dialog.open(ModaldialogComponent, {
     width: '600px',
   });
   this.appwindowWatcher = timer(500).pipe(repeat()).subscribe(() => this.watchAppwindow(this.appwindow,dialogRef));
   // this.appwindow.location.assign(windowloc);
 }

 public connect(job: Job, appinst?: any) {
   // this.statusMsg.next(null);
   let headers = new HttpHeaders();
   let options = { headers: headers, withCredentials: true};
   // this.busy.next(true);
   this.getInterface(job); // getInterface will subsequently called getAppInstance, which will call createTunnel, which will openAppWindow
 }




public postAgentData(keyCert: KeyCert) {
  let headers = new HttpHeaders();
  let options = { headers: headers, withCredentials: true};

  this.statusMsg.next("Authorising ...")
  let data = {'key': keyCert.key, 'cert': keyCert.cert};
  return this.http.post<any>(this.Base+'/sshagent',data,options)
                      .pipe(catchError(this.handleError))

}

public postMkDir(id: Identity, path: string, name: string ) {
  let headers = new HttpHeaders();
  let options = { headers: headers, withCredentials: true};
  let params = new URLSearchParams;
  params.set('identity',JSON.stringify(id));
  params.set('path',JSON.stringify(path));

  return this.http.post<any>(this.Base+'/mkdir?'+params.toString(),name, options);
}

public getSftpData(id: Identity, path: string, cd: string ) {
  let headers = new HttpHeaders();
  let options = { headers: headers, withCredentials: true};
  let params = new URLSearchParams;
  params.set('identity',JSON.stringify(id));
  params.set('path',JSON.stringify(path));
  params.set('cd',JSON.stringify(cd));

  return this.http.get<any>(this.Base+'/ls?'+params.toString(),options)
                      // .pipe(catchError(this.handleError))

}



public getstatusMsgSubject(): BehaviorSubject<any> {
  return this.statusMsg;
}

private httperror(errorstr: string) {
  var re = /login expired/gi;
  let searchresult = errorstr.search(re);
  if (searchresult != -1) {
    this.statusMsg.next("Some authentication tokens have expired, you may need to log in again");
    console.error(errorstr);
    this.authorisationService.updateAgentContents();
  }
  console.error(errorstr);
}

// private networkError(error: HttpErrorResponse) {
//   if (error.error instanceof ErrorEvent) {
//     console.log('network error contacting TES backend');
//   } else {
//     return error;
//   }
// }


private handleError(error: HttpErrorResponse) {
  console.error('in handleError');
  console.error(error);
  if (error.error instanceof ErrorEvent) {
    // A client-side or network error occurred. Handle it accordingly.
    return throwError("Hmm, that didn't work. If you're using a local connection, please make sure Strudel-TES is running.");
  } else {
    console.log('NOT an instance of ErrorEvent');
    console.log(error);
    // Not sure if this code works correctly. It should update identities in case the error is
    // that the user isn't allowed to run the job

    var re = /identity/gi;
    if (error.error.message != undefined) {
      let searchresult = error.error.message.search(re);
      if (searchresult != -1) {
        // this.getIdentities();
        return throwError('login expired, refreshing');
      }
      return throwError(error.error.message)
    } else {
      return throwError('An error occured but I\'m not sure exactly what. Please try again latter or contact the sys admins');
    }
    // this.statusMsg.next("There was an error submitting that job. The backend gave me the message: " + error.error.message);

  }
}



}
